
# Copyright (C) 2003  Pedro Werneck
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

"""Testing module for Loop module

"""

# System imports
import sys
if __name__ == "__main__":
    sys.path.insert(0, "..")

import time

# HoneyPython imports

from Honey.Loop import TaskQueue
from Honey.Loop import Engine

import unittest

class TestTaskQueue(unittest.TestCase):
    def setUp(self):
        self.queue = TaskQueue()

    def _queueCallback(self, *args, **kwds):
        """Helper method"""
        pass

    def testQueueAttributes(self):
        """Base attributes are sane"""
        self.assert_(hasattr(self.queue, "tasks"))

    def registerTask(self, delay=0):
        """Register a task"""
        return self.queue.put(delay, self._queueCallback, 1, 2, a=10, b=20)

    def testRegister(self):
        """Task registration"""
        self.registerTask()

    def testContains(self):
        """'in' keyword works properly"""
        id = self.registerTask()
        self.assert_(id in self.queue)
        
    def testRegisterInvalid(self):
        """Exception raises on invalid task registration"""
        self.assertRaises(TypeError, self.queue.put, 0, None)
        self.assertRaises(ValueError, self.queue.put, -1, type)

    def testDeadlines(self):
        """Deadlines are in the future"""
        now = time.time()
        deadline = self.registerTask()
        self.assert_(deadline > now)
        
    def testDeadlinesDoNotRepeat(self):
        """Two simultaneous tasks do not share deadlines"""
        id1 = self.registerTask()
        id2 = self.registerTask()
        self.assertNotEqual(id1, id2)

    def testUnregister(self):
        """Task unregistration"""
        id = self.registerTask()
        self.queue.remove(id)

    def testInvalidUnregister(self):
        """Exception is raised on unregistering when id does not exist"""
        self.assertRaises(KeyError, self.queue.remove, 1)

    def testGetTask(self):
        """Task is returned when requested"""
        id = self.registerTask()
        call, args, kwds = self.queue.get(id)
        self.assertEqual(call, self._queueCallback, "call")
        self.assertEqual(args, (1, 2), "args")
        self.assertEqual(kwds, {"a":10, "b":20}, "kwds")

    def testTaskRemoved(self):
        """Task is removed correctly"""
        id = self.registerTask()
        call, args, kwds = self.queue.get(id)
        self.assertEqual(call, self._queueCallback, "call")
        self.assertEqual(args, (1, 2), "args")
        self.assertEqual(kwds, {"a":10, "b":20}, "kwds")
        self.assertRaises(KeyError, self.queue.get, id)

    def testGetInvalidTask(self):
        """Exception is raised when getting invalid task"""
        self.assertRaises(KeyError, self.queue.get, 0)

    def testGetIterUpToNow(self):
        """Iterating over queue returns only expired tasks"""
        id1 = self.registerTask()
        id2 = self.registerTask()
        id3 = self.registerTask()
        id4 = self.registerTask(0.01)
        id5 = self.registerTask(1)
        id6 = self.registerTask(2)
        time.sleep(0.1)
        for task in self.queue:
            self.assertEqual(len(task), 3)
        self.assert_(id1 not in self.queue)
        self.assert_(id2 not in self.queue)
        self.assert_(id3 not in self.queue)
        self.assert_(id4 not in self.queue)
        self.assert_(id5 in self.queue)
        self.assert_(id6 in self.queue)

class TestEngine(unittest.TestCase):
    def setUp(self):
        self.engine = Engine()

    def testEngineAttributes(self):
        """Base attributes"""
        self.assert_(type(self.engine.queue) is TaskQueue)
        self.assert_(not self.engine.running)

    def _engineCallback(self, *args, **kwds):
        """Helper method"""
        self.assertEqual(args, (3,))
        self.assertEqual(kwds, {"a":10})

    def registerTask(self, delay=0):
        """Register a task"""
        return self.engine.register(delay, self._engineCallback, 3, a=10)

    def testRegister(self):
        """Task registration"""
        self.registerTask()

    def testRegisterInvalidCall(self):
        """Exception raises on invalid task registration"""
        self.assertRaises(TypeError, self.engine.register, 0, None)
        self.assertRaises(ValueError, self.engine.register, -1, type)

    def testDeadlines(self):
        """Deadlines are in the future"""
        now = time.time()
        deadline = self.registerTask()
        self.assert_(deadline > now)
        
    def testDeadlinesDoNotRepeat(self):
        """Two simultaneous tasks do not share deadlines"""
        id1 = self.registerTask()
        id2 = self.registerTask()
        self.assertNotEqual(id1, id2)

    def testUnregister(self):
        """Task unregistration"""
        id = self.registerTask()
        self.engine.unregister(id)

    def testInvalidUnregister(self):
        """Exception is raised when id does not exist"""
        self.assertRaises(KeyError, self.engine.unregister, 1)

    def _raiseCallback(self):
        """Helper method: always raise a weird exception"""
        raise ZeroDivisionError("I just raise a exception, nothing more")

    def _nullCallback(self):
        """Helper method: just mark that was called """
        self.called = True

    def _failCallback(self):
        """Helper method: fail imediately"""
        self.fail("Ops... this callback should never be called")

    def testTaskUpdating(self):
        """Update is working"""
        self.called = False
        self.engine.register(0, self._nullCallback)
        self.engine.register(0.1, self._raiseCallback)
        self.engine.register(1, self._failCallback)
        self.engine.update()
        self.assert_(self.called)
        time.sleep(0.2)
        self.assertRaises(ZeroDivisionError, self.engine.update)

    def testExceptionHandling(self):
        """Exceptions are being handled by the 'handleException' method """
        self.engine.register(0, self._raiseCallback)

        def handleException(exc, value=None, tb=None):
            self.assert_(exc is ZeroDivisionError)
            self.assert_(value is not None)
            self.assert_(tb is not None)

        self.engine.handleException = handleException
        self.engine.update()

    def _countCallback(self, *args, **kwds):
        """Helper method: register how many times it was called"""
        self.count += 1
        c = self.count
        self.assertEqual(args, (c,))
        self.assertEqual(kwds, {"a":c*10})
            
    def testEngineRunningAndStopping(self):
        """Engine is running and stopping"""
        self.count = 0
        self.engine.register(0, self._countCallback, 1, a=10)
        self.engine.register(0.02, self._countCallback, 3, a=30)
        self.engine.register(0.06, self._countCallback, 5, a=50)
        self.engine.register(0.04, self._countCallback, 4, a=40)
        self.engine.register(0.08, self._countCallback, 6, a=60)
        self.engine.register(1, self._countCallback, 7, a=70)
        self.engine.register(0, self._countCallback, 2, a=20)
        self.engine.register(0.5, self.engine.stop)
        self.engine.run()
        self.assertEqual(self.count, 6)
        


if __name__ == "__main__":
    sys.path.insert(0, "..")
    unittest.main()
    
